! compares 3 different algorithms in an Ax=b linear system resolution
! uses the A matrix stored in a Harwell-Boeing file, and outputs a log file
! with convergence and processing time
program msolve
  use misc_interface
  use solve_interface
  use csm_mod
  implicit none
  type(csmat) :: csm
  real(kind(0.d0)),allocatable,dimension(:) :: x,b,x_start
  real(kind(0.d0)) :: eps_max,eps_end
  integer :: i,n,i_max,i_end,dim_k
  character(len=80) :: filename

  logical :: imax_flag=.true.,eps_flag=.true.,dimk_flag=.true.
  logical :: fail_flag,info_flag=.false.,log_flag=.true.
  logical :: lanczos_flag=.false.,bicgstab_flag=.false.,gmres_flag=.false.
  character(len=40) :: logname
  character(len=20) :: logtag=''

  interface
     subroutine msolve_params(filename,logtag,n,i_max,imax_flag,&
          eps_max,eps_flag,dim_k,dimk_flag,info_flag,log_flag,&
          lanczos_flag,bicgstab_flag,gmres_flag)
       character(len=80),intent(in) :: filename
       character(len=20),intent(out) :: logtag
       real(kind(0.d0)),intent(out) :: eps_max
       integer,intent(out) :: i_max,n,dim_k
       logical,intent(out) :: imax_flag,eps_flag,dimk_flag
       logical,intent(out) :: info_flag,log_flag
       logical,intent(out) :: lanczos_flag,bicgstab_flag,gmres_flag
     endsubroutine msolve_params
     subroutine msolve_init(filename,csm,logname,logtag,log_flag,x,x_start,b)
       use csm_mod
       character(len=80),intent(in) :: filename
       type(csmat),intent(out) :: csm
       character(len=40),intent(out) :: logname
       character(len=20),intent(in) :: logtag
       logical,intent(in) :: log_flag
       real(kind(0.d0)),allocatable,dimension(:),intent(out) :: x,b,x_start
     endsubroutine msolve_init
     subroutine msolve_pre(mode,log_flag,logname,n,nval,i_max,eps_max,dim_k)
       character(len=*),intent(in) :: mode,logname
       logical,intent(in) :: log_flag
       integer,intent(in) :: n,nval,i_max,dim_k
       real(kind(0.d0)),intent(in) :: eps_max
     endsubroutine msolve_pre
     subroutine msolve_post(mode,filename,x,log_flag,info_flag,&
          eps_end,eps_max,i_end,i_max,fail_flag,n,nval)
       character(len=*),intent(in) :: mode,filename
       real(kind(0.d0)),dimension(:),intent(in) :: x
       logical,intent(in) :: log_flag,info_flag,fail_flag
       integer,intent(in) :: n,nval,i_max,i_end
       real(kind(0.d0)),intent(in) :: eps_max,eps_end
     endsubroutine msolve_post
  endinterface

  call msolve_params('msolve.params',logtag,n,i_max,&
       imax_flag,eps_max,eps_flag,dim_k,dimk_flag,info_flag,log_flag,&
       lanczos_flag,bicgstab_flag,gmres_flag)

  if(lanczos_flag.or.bicgstab_flag.or.gmres_flag) then
     if(eps_flag) then
        print *,'EPSILON :'
        read *,eps_max
     endif
     if(imax_flag) then
        print *,'MAX ITER :'
        read *,i_max
     endif
     if(gmres_flag.and.dimk_flag) then
        print *,'GMRES KRYLOV SPACE DIM :'
        read *,dim_k
     endif
  endif

  do i=1,n
     filename=get_param('msolve.params',i)
     if(filename(1:1)/='-') then

        call msolve_init(filename,csm,logname,logtag,log_flag,x,x_start,b)

! ------------------ lanczos -------------------

        if(lanczos_flag) then
           call msolve_pre('lanczos',log_flag,logname,&
                csm%n,csm%nval,i_max,eps_max,dim_k)
           x=x_start
           call solve_lanczos(a=csm,b=b,x=x,eps_max=eps_max,i_max=i_max,&
                eps_end=eps_end,i_end=i_end,log=log_flag)
           call msolve_post('lanczos',filename,x,log_flag,info_flag,&
                eps_end,eps_max,i_end,i_max,fail_flag,csm%n,csm%nval)
        endif

! ------------------ bicgstab ------------------

        if(bicgstab_flag) then
           call msolve_pre('bicgstab',log_flag,logname,&
                csm%n,csm%nval,i_max,eps_max,dim_k)
           x=x_start
           call solve_bicgstab(a=csm,b=b,x=x,eps_max=eps_max,i_max=i_max,&
                eps_end=eps_end,i_end=i_end,fail=fail_flag,log=log_flag)
           call msolve_post('bicgstab',filename,x,log_flag,info_flag,&
                eps_end,eps_max,i_end,i_max,fail_flag,csm%n,csm%nval)
        endif

! ------------------- gmres --------------------

        if(gmres_flag) then
           call msolve_pre('gmres',log_flag,logname,&
                csm%n,csm%nval,i_max,eps_max,dim_k)
           x=x_start
           call solve_gmres(a=csm,b=b,x=x,eps_max=eps_max,i_max=i_max,&
                eps_end=eps_end,i_end=i_end,dim_k=dim_k,&
                fail=fail_flag,log=log_flag)
           call msolve_post('gmres',filename,x,log_flag,info_flag,&
                eps_end,eps_max,i_end,i_max,fail_flag,csm%n,csm%nval)
        endif
     endif
  enddo
endprogram msolve
