! reads the main matrix from the Harwell-Boeing file, modifies this
! matrix to ensure correct results, and initializes x, b=Ax, and x_start
subroutine msolve_init(filename,csm,logname,logtag,log_flag,x,x_start,b)
  use misc_interface
  use csm_interface
  use csm_mod
  implicit none
  character(len=80),intent(in) :: filename
  type(csmat),intent(out) :: csm
  character(len=40),intent(out) :: logname
  character(len=20),intent(in) :: logtag
  logical,intent(in) :: log_flag
  real(kind(0.d0)),allocatable,dimension(:),intent(out) :: x,b,x_start

  real(kind(0.d0)),allocatable,dimension(:,:) :: a
  real(kind(0.d0)) :: maxi
  integer :: i,j

  call csm_load(csm,filename)
  maxi=max(maxval(csm%values),-minval(csm%values))
  call csm2mat(csm,a) ! very heavy in memory,
                      ! but other methods, using the sparse matrix structure,
                      ! always resulted in segmentation faults
                      ! -> needs to be rewritten
  do j=1,csm%n
     a(j,j)=a(j,j)+maxi
  enddo
  call mat2csm(a,csm)

  if(log_flag) then
     logname='-'//filename(index(filename,'/',back=.true.)+1:&
       index(filename,'.',back=.true.)-1)
     if(trim(logtag)/='') logname=trim(logname)//'.'//trim(logtag)
  endif
  call d_vect_alloc(x,csm%n)
  call d_vect_alloc(b,csm%n)
  do i=1,csm%n
     x(i)=i
  enddo
  call csm_vect_pdt(csm,x,b)
  call d_vect_alloc(x_start,csm%n)
  x_start=1.5*x
  do i=1,csm%n,2
     x_start(i)=x_start(i)/3
  enddo
return
endsubroutine msolve_init
