! solves the AX=B system, X unknown, with the BiCGStab method
! with the provided X as guess values,
! eps_max as tolerance, i_max as max iteration
subroutine solve_bicgstab(a,b,x,eps_max,i_max,eps_end,i_end,fail,log)
  use csm_mod
  use misc_interface
  use csm_interface
  implicit none
  type(csmat),intent(in) :: a
  real(kind(0.d0)),dimension(:),intent(in) :: b
  real(kind(0.d0)),dimension(:),intent(inout) :: x
  real(kind(0.d0)),intent(in) :: eps_max
  integer,intent(in) :: i_max
  real(kind(0.d0)),optional,intent(out) :: eps_end
  integer,optional,intent(out) :: i_end
  logical,optional,intent(out) :: fail
  logical,optional,intent(in) :: log

  real(kind(0.d0)),allocatable,dimension(:) :: r,p,u,y,a_p,a_u,a_x
  real(kind(0.d0)) :: alpha,beta,gamma,rho,rhobar

  real(kind(0.d0)) :: eps
  integer :: i
  real :: start_time,time
  logical :: log_flag=.false.,fail_flag=.false.

  if((a%n/=size(b)).or.(a%n/=size(x))) call error(409)
  if(present(log)) log_flag=log


  call d_vect_alloc(r,a%n)
  call d_vect_alloc(p,a%n)
  call d_vect_alloc(y,a%n)
  call d_vect_alloc(u,a%n)
  call d_vect_alloc(a_p,a%n)
  call d_vect_alloc(a_u,a%n)
  call d_vect_alloc(a_x,a%n)

  fail_flag=.false.
  call csm_vect_pdt(a,x,a_x)
  r=b-a_x
  y=r

  i=0
  if(log_flag) call cpu_time(start_time)
  do
     eps=rel_diff(a_x,b)
     if(log_flag) then
        call cpu_time(time)
        call log_add(trim(adjustl(int2char(i)))//&
             ' '//trim(adjustl(real2char(time-start_time)))//&
             ' '//trim(adjustl(double2char(eps))))
     endif
     if((fail_flag).or.(i>=i_max).or.(eps<=eps_max)) exit

     rhobar=rho
     rho=dot_product(y,r)
     if(rho==0) then
        fail_flag=.true.
     else
        if(i==0) then
           p=r
        else
           alpha=(rho/rhobar)*(beta/gamma)
           p=r+alpha*(p-gamma*a_p)
        endif
        call csm_vect_pdt(a,p,a_p)
        beta=rho/dot_product(y,a_p)
        u=r-beta*a_p
        call csm_vect_pdt(a,u,a_u)
        gamma=dot_product(u,a_u)/dot_product(a_u,a_u)
        x=x+beta*p+gamma*u
        r=u-gamma*A_u
     endif

     call csm_vect_pdt(a,x,a_x)
     i=i+1
  enddo

  fail=fail_flag
  eps_end=eps
  i_end=i
  
endsubroutine solve_bicgstab
