! solves the AX=B system, X unknown, with the restarted GMRES method
! with the provided X as guess values,
! eps_max as tolerance, i_max as max iteration, dim_k as the krylov space dim
subroutine solve_gmres(a,b,x,eps_max,i_max,eps_end,i_end,dim_k,fail,log)
  use csm_mod
  use misc_interface
  use csm_interface
  implicit none
  type(csmat),intent(in) :: a
  real(kind(0.d0)),dimension(:),intent(in) :: b
  real(kind(0.d0)),dimension(:),intent(inout) :: x
  real(kind(0.d0)),intent(in) :: eps_max
  integer,intent(in) :: i_max,dim_k
  real(kind(0.d0)),optional,intent(out) :: eps_end
  integer,optional,intent(out) :: i_end
  logical,optional,intent(out) :: fail
  logical,optional,intent(in) :: log

  real(kind(0.d0)) :: beta,alpha,c_alpha,s_alpha,hjk,hj1k
  real(kind(0.d0)),allocatable,dimension(:,:) :: v,h
  real(kind(0.d0)),allocatable,dimension(:) :: r,omega,y,a_x,xbar

  real(kind(0.d0)) :: eps,delta
  integer :: i,j,k,m
  real :: start_time,time
  logical :: log_flag=.false.

  if((a%n/=size(b)).or.(a%n/=size(x))) call error(409)
  if(present(log)) log_flag=log
  m=max(dim_k,2)
  m=min(m,a%n-1)

  call d_vect_alloc(r,a%n)
  call d_vect_alloc(xbar,a%n)
  call d_vect_alloc(omega,a%n)
  call d_vect_alloc(y,m)
  call d_mat_alloc(v,a%n,m)
  call d_mat_alloc(h,a%n,m)
  call d_vect_alloc(a_x,a%n)

  call csm_vect_pdt(a,x,a_x)

  h=0
  r=b-a_x
  beta = sqrt(dot_product(r,r))
  xbar=x+1

  i=0
  if(log_flag) call cpu_time(start_time)
  do

     eps=rel_diff(a_x,b)
     if(log_flag) then
        call cpu_time(time)
        call log_add(trim(adjustl(int2char(i)))//&
             ' '//trim(adjustl(real2char(time-start_time)))//&
             ' '//trim(adjustl(double2char(eps))))
     endif

     delta=max(maxval(x-xbar),-minval(x-xbar))
     if((delta==0).or.(i>=i_max).or.(eps<=eps_max)) exit
     xbar=x

     ! arnoldi method

     v(:,1)=r/beta
     do j=1,m
        call csm_vect_pdt(a,v(:,j),omega)
        do k=1,j
           h(k,j)=dot_product(omega,v(:,k))
           omega=omega-h(k,j)*v(:,k)
        enddo
        h(j+1,j)=sqrt(dot_product(omega,omega))
        if(h(j+1,j)==0) exit
        if(j<m) v(:,j+1)=omega/h(j+1,j)
     enddo

     ! triangulation of h and y update
     y=0
     y(1)=beta
     do j=1,m
        alpha=sqrt(h(j,j)*h(j,j)+h(j+1,j)*h(j+1,j))
        if(alpha==0) alpha=1
        c_alpha=h(j,j)/alpha
        s_alpha=-h(j+1,j)/alpha
        do k=1,m
           hjk=h(j,k)
           hj1k=h(j+1,k)
           h(j,k)=c_alpha*hjk-s_alpha*hj1k
           h(j+1,k)=s_alpha*hjk+c_alpha*hj1k
        enddo
        if(j<m) y(j+1)=s_alpha*y(j)
        y(j)=c_alpha*y(j)
     enddo

     ! solve hx=y->y
     do j=m,1,-1
        if(h(j,j)==0) h(j,j)=1
        y(j)=y(j)/h(j,j)
        y(1:j-1)=y(1:j-1)-y(j)*h(1:j-1,j)
     enddo

     x=x+matmul(v,y)

     call csm_vect_pdt(a,x,a_x)
     r=b-a_x
     beta = sqrt(dot_product(r,r))

     i=i+1
  enddo

  fail=(delta==0)
  eps_end=eps
  i_end=i
endsubroutine solve_gmres
