! solves the AX=B system, X unknown, with the Lanczos-Orthomin method
! with the provided X as guess values,
! eps_max as tolerance, i_max as max iteration
subroutine solve_lanczos(a,b,x,eps_max,i_max,eps_end,i_end,log)
  use csm_mod
  use misc_interface
  use csm_interface
  implicit none
  type(csmat),intent(in) :: a
  real(kind(0.d0)),dimension(:),intent(in) :: b
  real(kind(0.d0)),dimension(:),intent(inout) :: x
  real(kind(0.d0)),intent(in) :: eps_max
  integer,intent(in) :: i_max
  real(kind(0.d0)),optional,intent(out) :: eps_end
  integer,optional,intent(out) :: i_end
  logical,optional,intent(in) :: log

  real(kind(0.d0)),allocatable,dimension(:) :: r,rbar,p,pbar,a_p,a_x,pbar_a
  real(kind(0.d0)) :: alpha,beta,rbar_r

  real(kind(0.d0)) :: eps
  integer :: i
  real :: start_time,time
  logical :: log_flag=.false.

  if((a%n/=size(b)).or.(a%n/=size(x))) call error(409)
  if(present(log)) log_flag=log

  call d_vect_alloc(r,a%n)
  call d_vect_alloc(p,a%n)
  call d_vect_alloc(rbar,a%n)
  call d_vect_alloc(pbar,a%n)
  call d_vect_alloc(a_x,a%n)
  call d_vect_alloc(a_p,a%n)
  call d_vect_alloc(pbar_a,a%n)

  call csm_vect_pdt(a,x,a_x)
  r=b-a_x
  p=r
  rbar=r
  pbar=rbar

  i=0
  if(log_flag) call cpu_time(start_time)
  do
     eps=rel_diff(a_x,b)
     if(log_flag) then
        call cpu_time(time)
        call log_add(trim(adjustl(int2char(i)))//&
             ' '//trim(adjustl(real2char(time-start_time)))//&
             ' '//trim(adjustl(double2char(eps))))
     endif
     if((i>=i_max).or.(eps<=eps_max)) exit

     call csm_vect_pdt(a,p,a_p)
     rbar_r=dot_product(rbar,r)
     beta=rbar_r/dot_product(pbar,a_p)
     r=r-beta*a_p
     x=x+beta*p
     call vect_csm_pdt(pbar,a,pbar_a)
     rbar=rbar-beta*pbar_a
     alpha=dot_product(rbar,r)/rbar_r
     p=r+alpha*p
     pbar=rbar+alpha*pbar

     call csm_vect_pdt(a,x,a_x)
     i=i+1
  enddo

  eps_end=eps
  i_end=i

endsubroutine solve_lanczos
